/* SPDX-License-Identifier: GPL-2.0-or-later */
/*
 *  Copyright (C) 2020 Namjae Jeon <linkinjeon@kernel.org>
 */

#ifndef _VERSION_H

#define EXFAT_PROGS_VERSION "1.4.0"
#define EXFAT_PROGS_RELEASE_DATE "2026-05-26"

#endif /* !_VERSION_H */
